# This file was auto-generated by Fern from our API Definition.

import typing

ProviderApiKeyWriteProvider = typing.Union[
    typing.Literal["openai", "anthropic", "gemini", "openrouter", "vertex-ai", "bedrock", "custom-llm", "opik-free"],
    typing.Any,
]
