# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata


class DatasetVersionDiffStats(UniversalBaseModel):
    items_added: typing_extensions.Annotated[typing.Optional[int], FieldMetadata(alias="itemsAdded")] = None
    items_modified: typing_extensions.Annotated[typing.Optional[int], FieldMetadata(alias="itemsModified")] = None
    items_deleted: typing_extensions.Annotated[typing.Optional[int], FieldMetadata(alias="itemsDeleted")] = None
    items_unchanged: typing_extensions.Annotated[typing.Optional[int], FieldMetadata(alias="itemsUnchanged")] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
