# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata


class TraceEnrichmentOptions(UniversalBaseModel):
    """
    Options for enriching trace data
    """

    include_spans: typing_extensions.Annotated[typing.Optional[bool], FieldMetadata(alias="includeSpans")] = None
    include_tags: typing_extensions.Annotated[typing.Optional[bool], FieldMetadata(alias="includeTags")] = None
    include_feedback_scores: typing_extensions.Annotated[
        typing.Optional[bool], FieldMetadata(alias="includeFeedbackScores")
    ] = None
    include_comments: typing_extensions.Annotated[typing.Optional[bool], FieldMetadata(alias="includeComments")] = None
    include_usage: typing_extensions.Annotated[typing.Optional[bool], FieldMetadata(alias="includeUsage")] = None
    include_metadata: typing_extensions.Annotated[typing.Optional[bool], FieldMetadata(alias="includeMetadata")] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
