# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class DatasetVersionSummary(UniversalBaseModel):
    """
    Summary of the latest dataset version
    """

    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Unique identifier of the version
    """

    version_hash: typing.Optional[str] = pydantic.Field(default=None)
    """
    Hash of the version content
    """

    version_name: typing.Optional[str] = pydantic.Field(default=None)
    """
    Sequential version name formatted as 'v1', 'v2', etc.
    """

    change_description: typing.Optional[str] = pydantic.Field(default=None)
    """
    Description of changes in this version
    """

    tags: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Tags associated with this version
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
