# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class ManualEvaluationResponse(UniversalBaseModel):
    entities_queued: typing.Optional[int] = pydantic.Field(default=None)
    """
    Number of entities queued for evaluation
    """

    rules_applied: typing.Optional[int] = pydantic.Field(default=None)
    """
    Number of rules that will be applied
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
